from six.moves.urllib.parse import urlparse, urlencode
import re
import logging

from owmeta_core.graph_object import IdentifierMissingException
from owmeta_core.context import Context
import owmeta_core.dataobject_property as DP
from owmeta_core.dataobject import DataObject, DatatypeProperty, Alias, BaseDataObject
import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

from . import SCI_CTX
from . import bibtex as BIB


logger = logging.getLogger(__name__)


class WormbaseRetrievalException(Exception):
    pass


class PubmedRetrievalException(Exception):
    pass


# A little bit about why this a separate type from Document:
#
# This type corresponds to a document which has some statements that we care
# about. The key reason this is distinct from Document is that a document need
# not provide evidence of anything. For example, the `WormData.n4` file
# generated by insert_worm.py is a document, but it doesn't provide any
# scientific or logical justification for any of the statements made within it.
class BaseDocument(DataObject):
    class_context = SCI_CTX

    def make_context_identifier(self):
        return self.make_identifier(self.identifier)

    @property
    def as_context(self):
        if self.context is not None:
            return Context.contextualize(self.context)(ident=self.make_context_identifier())
        else:
            return Context(ident=self.make_context_identifier())


class Document(BaseDocument):
    """
    A representation of some document.

    Possible keys include::

        pmid, pubmed: a pubmed id or url (e.g., 24098140)
        wbid, wormbase: a wormbase id or url (e.g., WBPaper00044287)
        doi: a Digitial Object id or url (e.g., s00454-010-9273-0)
        uri: a URI specific to the document, preferably usable for accessing
             the document
    """

    class_context = SCI_CTX

    author = DatatypeProperty(multiple=True)
    ''' An author of the document '''

    doi = DatatypeProperty()
    ''' A Digital Object Identifier (DOI), optional '''

    uri = DatatypeProperty(multiple=True)
    ''' A non-standard URI for the document '''

    wbid = DatatypeProperty()
    ''' An ID from WormBase.org that points to a record, optional '''

    wormbaseid = Alias(wbid)
    ''' An alias to `wbid` '''

    pmid = DatatypeProperty()
    ''' A PubMed ID (PMID) that points to a paper '''

    year = DatatypeProperty()
    ''' The year (e.g., publication year) of the document '''

    date = Alias(year)
    ''' Alias to year '''

    title = DatatypeProperty()
    ''' The title of the document '''

    def __init__(
            self,
            bibtex=None,
            doi=None,
            pubmed=None,
            wormbase=None,
            **kwargs):
        """
        Parameters
        ----------
        bibtex : string
            A string containing a single BibTeX entry. Parsed during initialization, but not saved thereafter. optional
        doi : string
            A Digital Object Identifier (DOI). optional
        pubmed : string
            A PubMed ID (PMID) or URL that points to a paper. Ignored if 'pmid' is provided. optional
        wormbase : string
            An ID or URL from WormBase that points to a record. Ignored if `wbid` or `wormbaseid` are provided. optional
        """
        super(Document, self).__init__(**kwargs)

        self.id_precedence = ('doi', 'pmid', 'wbid', 'uri')

        if bibtex is not None:
            self.update_with_bibtex(bibtex)

        if pubmed is not None and not self.pmid.has_defined_value():
            if pubmed[:4] == 'http':
                _tmp = _pubmed_uri_to_pmid(pubmed)
                if _tmp is None:
                    raise ValueError("Couldn't convert Pubmed URL to a PubMed ID")
                pmid = _tmp
            else:
                pmid = pubmed
            self.pmid.set(pmid)

        if wormbase is not None and not self.wbid.has_defined_value():
            if wormbase[:4] == 'http':
                _tmp = _wormbase_uri_to_wbid(wormbase)
                if _tmp is None:
                    raise ValueError("Couldn't convert Wormbase URL to a Wormbase ID")
                wbid = _tmp
            else:
                wbid = wormbase
            self.wbid.set(wbid)

        if doi is not None:
            if doi[:4] == 'http':
                _tmp = _doi_uri_to_doi(doi)
                if _tmp is not None:
                    doi = _tmp
            self.doi.set(doi)

    def update_with_bibtex(self, bibtex):
        bib_db = BIB.loads(bibtex)
        if len(bib_db.entries) > 1:
            raise ValueError('The given BibTex string has %d entries.'
                             ' Cannot determine which entry to use for the document' % len(bib_db))
        BIB.update_document_with_bibtex(self, bib_db.entries[0])

    def defined_augment(self):
        for x in self.id_precedence:
            if getattr(self, x).has_defined_value():
                return True
        return False

    def identifier_augment(self):
        for idKind in self.id_precedence:
            idprop = getattr(self, idKind)
            if idprop.has_defined_value():
                s = str(idKind) + ":" + idprop.defined_values[0].identifier.n3()
                return self.make_identifier(s)
        raise IdentifierMissingException(self)

    def update_from_wormbase(self, replace_existing=False, **kwargs):
        """ Queries WormBase.org for additional data to fill in the `Document`.

        If replace_existing is set to `True`, then existing values will be cleared.

        Parameters
        ----------
        replace_existing : bool
            Whether to replace values that are already set for a given property
        **kwargs
            Passed on as arguments to `requests.Session.get`
        """

        # XXX: wormbase's REST API is pretty sparse in terms of data provided.
        #     Would be better off using AQL or the perl interface
        # _Very_ few of these have these fields filled in
        wbid = self.wbid.defined_values
        if len(wbid) == 1:
            wbid = wbid[0].identifier.toPython()

            # get the author
            try:
                root = self.conf.get('wormbase_api_root_url', 'http://rest.wormbase.org')
                url = f'{root}/rest/widget/paper/{wbid}/overview?content-type=application%2Fjson'
                j = _json_request(url, **kwargs)
                if 'fields' in j:
                    f = j['fields']
                    if 'authors' in f:
                        dat = f['authors']['data']
                        if dat is not None:
                            if replace_existing and self.author.has_defined_value:
                                self.author.clear()
                            for x in dat:
                                self.author.set(x['label'])

                    for fname in ('pmid', 'year', 'title', 'doi'):
                        if fname in f and f[fname]['data'] is not None:
                            attr = getattr(self, fname)
                            if replace_existing and attr.has_defined_value:
                                attr.clear()
                            attr.set(f[fname]['data'])
            except Exception:
                logger.warning("Couldn't retrieve Wormbase data", exc_info=True)
        elif len(wbid) == 0:
            raise WormbaseRetrievalException("There is no Wormbase ID attached to this Document."
                                             " So no data can be retrieved")
        else:
            raise WormbaseRetrievalException("There is more than one Wormbase ID attached to this Document."
                                             " Please try with just one Wormbase ID")

    def _crossref_doi_extract(self):
        # Extract data from crossref
        def crRequest(doi):
            data = {'q': doi}
            data_encoded = urlencode(data)
            return _json_request(
                'http://search.labs.crossref.org/dois?%s' %
                data_encoded)

        doi = self.doi()
        if doi[:4] == 'http':
            doi = _doi_uri_to_doi(doi)
        try:
            r = crRequest(doi)
        except Exception:
            logger.warning("Couldn't retrieve Crossref info", exc_info=True)
            return
        # XXX: I don't think coins is meant to be used, but it has structured
        # data...
        if len(r) > 0:
            extra_data = r[0]['coins'].split('&amp;')
            fields = (x.split("=") for x in extra_data)
            fields = [[y.replace('+', ' ').strip() for y in x] for x in fields]
            authors = [x[1] for x in fields if x[0] == 'rft.au']
            for a in authors:
                self.author(a)
            # no error for bad ids, just an empty list
            if len(r) > 0:
                # Crossref can process multiple doi's at one go and return the
                # metadata. we just need the first one
                r = r[0]
                if 'title' in r:
                    self.title(r['title'])
                if 'year' in r:
                    self.year(r['year'])

    def update_from_pubmed(self, read_size=2**16, **kwargs):
        '''
        Update the document attributes from NCBI Entrez API using the pubmed attribute

        Parameters
        ----------
        chunk_size : int
            The number of bytes to pass to `requests.Response.iter_content`. This *may*
            reduce runtime memory requirements for the request.
        **kwargs
            Passed on as arguments to `requests.Session.get`
        '''

        def pmRequest(pmid):
            import xml.etree.ElementTree as ET  # Python 2.5 and up

            url = ('https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esummary.fcgi?'
                    f'db=pubmed&id={pmid}')
            key = self.get('pubmed.api_key', None)
            if key:
                url += f'&api_key={key}'
            else:
                logger.warning("PubMed API key not defined. API calls will be limited.")

            if 'do_retries' not in kwargs:
                kwargs['do_retries'] = True

            kwargs['stream'] = True

            s = _url_request(url, **kwargs)
            if hasattr(s, 'charset'):
                parser = ET.XMLParser(encoding=s.charset)
            else:
                parser = ET.XMLParser(encoding='UTF-8')

            with s:
                for chunk in s.iter_content(read_size):
                    parser.feed(chunk)
                return parser.close()

        pmid = self.pmid.defined_values
        if len(pmid) == 1:
            pmid = pmid[0].identifier.toPython()
            try:
                tree = pmRequest(pmid)
            except Exception:
                logger.warning("Couldn't retrieve Pubmed info", exc_info=True)
                return
            for x in tree.findall('./DocSum/Item[@Name="AuthorList"]/Item'):
                self.author(x.text)

            for x in tree.findall('./DocSum/Item[@Name="Title"]'):
                self.title(x.text)

            for x in tree.findall('./DocSum/Item[@Name="DOI"]'):
                self.doi(x.text)

            for x in tree.findall('./DocSum/Item[@Name="PubDate"]'):
                self.year(x.text)

        elif len(pmid) == 0:
            raise PubmedRetrievalException('No Pubmed ID is attached to this document. Cannot retrieve Pubmed data')
        else:
            raise PubmedRetrievalException('More than one Pubmed ID is attached to this document.'
                                           ' Please try with just one Pubmed ID')


class SourcedFrom(DP.ObjectProperty):
    '''
    Indicates which document provided the source for an object
    '''
    class_context = SCI_CTX
    link_name = "sourced_from"
    value_type = BaseDocument
    owner_type = BaseDataObject
    multiple = False
    lazy = True


def _wormbase_uri_to_wbid(uri):
    return str(urlparse(uri).path.split("/")[2])


def _pubmed_uri_to_pmid(uri):
    return str(urlparse(uri).path.split("/")[2])


def _doi_uri_to_doi(uri):
    # DOI URL to DOI translation is complicated. This is a cop-out.
    parsed = urlparse(uri)
    if 'doi.org' in parsed.netloc:
        doi = parsed.path.split("/", 1)[1]
    else:
        doi = None

    return doi


def _url_request(url, requests_session=None, do_retries=False, **kwargs):

    if requests_session is None:
        sess = requests.Session()
    else:
        sess = requests_session

    if do_retries:
        retries = Retry()
        adapter = HTTPAdapter(max_retries=retries)
        sess.mount('http://', adapter)
        sess.mount('https://', adapter)

    if 'timeout' not in kwargs:
        kwargs['timeout'] = 1

    try:
        resp = sess.get(url, **kwargs)
        if resp.status_code != 200:
            raise Exception(f'Service returned status code {resp.status_code}')
        content_type = resp.headers.get('content-type')
        if content_type:
            md = re.search("charset *= *([^ ]+)", content_type)
            if md:
                resp.charset = md.group(1)

        return resp
    except Exception:
        logger.error("Error in request for %s", url, exc_info=True)
        raise


def _json_request(url, **kwargs):
    if 'headers' in kwargs:
        headers = kwargs['headers']
    else:
        headers = {}
        kwargs['headers'] = headers
    headers['Accept'] = 'application/json'
    try:
        return _url_request(url, **kwargs).json()
    except BaseException:
        logger.warning("Couldn't retrieve JSON data from %s", url,
                       exc_info=True)
        return {}



from owmeta_core.dataobject import DataObject, DatatypeProperty

from . import SCI_CTX


class Plot(DataObject):
    """
    Object for storing plot data in owmeta.

    Parameters
    ----------

    data : 2D list (list of lists)
        List of XY coordinates for this Plot.

    Example usage ::
        >>> pl = Plot([[1, 2], [3, 4]])
        >>> pl.get_data()
        # [[1, 2], [3, 4]]
    """

    class_context = SCI_CTX

    _data_string = DatatypeProperty()

    def __init__(self, data=None, *args, **kwargs):
        super(Plot, self).__init__(*args, **kwargs)

        if data:
            self.set_data(data)

    def _to_string(self, input_list):
        """
        Converts input_list to a string
        for serialized storage in owmeta.
        """
        return '|'.join([str(item) for item in input_list])

    def _to_list(self, input_string):
        """
        Converts from internal serlialized string
        to a 2D list.
        """
        if input_string is None:
            return None

        out_list = []
        for pair_string in input_string.split('|'):
            pair_as_list = pair_string \
                .replace('[', '') \
                .replace(']', '') \
                .split(',')
            out_list.append([float(x) for x in pair_as_list])
        return out_list

    def set_data(self, data):
        """
        Set the data attribute, which is user-facing,
        as well as the serialized _data_string
        attribute, which is used for db storage.
        """
        try:
            # make sure we're dealing with a 2D list
            assert isinstance(data, list)
            assert isinstance(data[0], list)
            self._data_string(self._to_string(data))
            self.data = data
        except (AssertionError, IndexError):
            raise ValueError('Attribute "data" must be a 2D list of numbers.')

    def get_data(self):
        """
        Get the data stored for this plot.
        """
        return self._to_list(self._data_string())



'''
`bibtexparser` customizations
'''
import re


HOWPUB_URL_RE = re.compile(r'\\url{([^}]+)}')


def customizations(record):
    """
    Standard owmeta `bibtexparser` customizations

    Includes: `url`, `note_url`, `doi`, `listify`, and `author`


    Parameters
    ----------
    record : dict
        the record

    Returns
    -------
    dict
        the given `record` with any updates applied
    """
    return url(note_url(doi(listify(author(record)))))


def listify_one(record, name):
    '''
    If the given field `name` does not have a `list` value, then updates the record by
    turning that value into a list.

    Parameters
    ----------
    record : dict
        The record to update
    name : str
        The name of the field to turn into a list

    Returns
    -------
    dict
        the given `record` with any updates applied
    '''
    if not isinstance(record[name], (list, tuple)):
        record[name] = [record[name]]
    elif isinstance(record[name], tuple):
        record[name] = list(record[name])
    return record


def listify(record):
    '''
    Turns every value in the record into a list except for ``ENTRYTYPE`` and ``ID``
    '''
    # Since some items can be multiples, it simplifies code in most places to
    # just make everything a list, even if it cannot appear more than once in
    # the properly formatted record.
    for val in record:
        if val not in ('ID', 'ENTRYTYPE'):
            listify_one(record, val)
    return record


def doi(record):
    """
    Adds a doi URI to the record if there's a ``doi`` entry in the record

    Parameters
    ----------
    record : dict
        the record to update

    Returns
    -------
    dict
        the given `record` with any updates applied
    """
    doi = record.get('doi')
    if doi is not None:
        if 'link' not in record:
            record['link'] = []
        for item in record['link']:
            if 'doi' in item:
                break
        else: # no break
            if not isinstance(doi, (list, tuple)):
                doi = [doi]

            for link in doi:
                if link.startswith('10'):
                    link = 'http://dx.doi.org/' + link
                record['link'].append(link)
    return record


def author(record):
    """
    Split author field by the string 'and' into a list of names.

    Parameters
    ----------
    record : dict
        the record

    Returns
    -------
    dict
        the given `record` with any updates applied
    """
    if "author" in record:
        if record["author"]:
            record["author"] = [i.strip() for i in record["author"].replace('\n', ' ').split(" and ")]
        else:
            del record["author"]
    return record


def note_url(record):
    '''
    Extracts URLs from ``note`` entries in the given record

    Parameters
    ----------
    record : dict
        the record

    Returns
    -------
    dict
        the given `record` with any updates applied
    '''
    note = record.get('note')
    if note is not None:
        for n in note:
            for u in HOWPUB_URL_RE.finditer(n):
                url = record.get('url')
                if url is None:
                    record['url'] = [u.group(1)]
                else:
                    listify_one(record, 'url')['url'].append(u.group(1))
    return record


def url(record):
    r'''
    Merges any URL from ``\url{...}`` in ``howpublished``, and any existing ``link`` or
    ``url`` values in the record and normalizes them into a `list` in the ``url`` field of
    the record

    Parameters
    ----------
    record : dict
        the record

    Returns
    -------
    dict
        the given `record` with any updates applied
    '''

    u = record.get('howpublished', '')
    md = HOWPUB_URL_RE.match(u)
    if md:
        v = record.get('url')
        if isinstance(v, tuple):
            record['url'] = list(v)

        if isinstance(v, list):
            v.append(md[1])

    url = record.get('url')
    link = record.get('link')

    if url is None:
        if isinstance(link, tuple):
            record['url'] = list(link)
        elif isinstance(link, list):
            record['url'] = link
        elif link is not None:
            record['url'] = [link]
        return record

    if isinstance(url, tuple):
        url = list(url)
        record['url'] = url

    if isinstance(url, list):
        if isinstance(link, (list, tuple)):
            url.extend(link)
        elif link is not None:
            url.append(link)

    return record



from .data_trans.neuron_data import NeuronCSVDataTranslator
from .data_trans.connections import (NeuronConnectomeCSVTranslator,
                                     NeuronConnectomeSynapseClassTranslator)
from .data_trans.wormatlas import WormAtlasCellListDataTranslator
from .data_trans.wormbase import (WormbaseTextMatchCSVTranslator,
                                  WormbaseIonChannelCSVTranslator,
                                  CellWormBaseCSVTranslator)
from .data_trans.context_merge import ContextMergeDataTranslator


def owm_data(ns):
    ctx = ns.context
    ctx.add_import(NeuronCSVDataTranslator.definition_context)
    ctx.add_import(ContextMergeDataTranslator.definition_context)
    ctx(NeuronConnectomeCSVTranslator)()
    ctx(NeuronConnectomeSynapseClassTranslator)()
    ctx(NeuronCSVDataTranslator)()
    ctx(WormAtlasCellListDataTranslator)()
    ctx(WormbaseIonChannelCSVTranslator)()
    ctx(WormbaseTextMatchCSVTranslator)()
    ctx(ContextMergeDataTranslator)()
    ctx(CellWormBaseCSVTranslator)()



import logging

from owmeta_core.dataobject import DataObject, ObjectProperty
from owmeta_core.context_dataobject import ContextDataObject
from owmeta_core.context import Context

from . import SCI_CTX

logger = logging.getLogger(__name__)


class EvidenceError(Exception):
    pass


class ContextToDataObjectMixin(object):
    def set(self, value):
        v = value
        if isinstance(value, Context):
            v = value.rdf_object
        return super(ContextToDataObjectMixin, self).set(v)


class Evidence(DataObject):
    """
    A representation which provides evidence, for a group of statements.

    Attaching evidence to an set of statements is done like this::

       >>> from owmeta.connection import Connection
       >>> from owmeta.evidence import Evidence
       >>> from owmeta_core.context import Context

    Declare contexts::

       >>> ACTX = Context(ident="http://example.org/data/some_statements")
       >>> BCTX = Context(ident="http://example.org/data/some_other_statements")
       >>> EVCTX = Context(ident="http://example.org/data/some_statements#evidence")

    Make statements in `ACTX` and `BCTX` contexts::

       >>> ACTX(Connection)(pre_cell="VA11", post_cell="VD12", number=3)
       >>> BCTX(Connection)(pre_cell="VA11", post_cell="VD12", number=2)

    In `EVCTX`, state that a that a certain document supports the set of
    statements in `ACTX`, but refutes the set of statements in `BCTX`::

       >>> doc = EVCTX(Document)(author='White et al.', date='1986')
       >>> EVCTX(Evidence)(reference=doc, supports=ACTX.rdf_object)
       >>> EVCTX(Evidence)(reference=doc, refutes=BCTX.rdf_object)

    Finally, save the contexts::

       >>> ACTX.save_context()
       >>> BCTX.save_context()
       >>> EVCTX.save_context()

    One note about the `reference` predicate: the reference should, ideally, be
    an unambiguous link to a peer-reviewed piece of scientific literature
    detailing methods and data analysis that supports the set of statements.
    However, in gather data from pre-existing sources, going to that level of
    specificity may be difficult due to deficient query capability at the data
    source. In such cases, a broader reference, such as a `Website` with
    information which guides readers to a peer-reviewed article supporting the
    statement is sufficient.

    """

    class_context = SCI_CTX

    supports = ObjectProperty(value_type=ContextDataObject,
                              mixins=(ContextToDataObjectMixin,))
    '''A context naming a set of statements which are supported by the attached
       reference'''

    refutes = ObjectProperty(value_type=ContextDataObject,
                             mixins=(ContextToDataObjectMixin,))
    '''A context naming a set of statements which are refuted by the attached
       reference'''

    reference = ObjectProperty()
    '''The resource providing evidence supporting/refuting the attached context'''

    def defined_augment(self):
        return ((self.supports.has_defined_value() or
                 self.refutes.has_defined_value()) and
                self.reference.has_defined_value())

    def identifier_augment(self):
        s = ""
        if self.supports.has_defined_value:
            s += self.supports.onedef().identifier.n3()
        else:
            s += self.refutes.onedef().identifier.n3()

        s += self.reference.onedef().identifier.n3()
        return self.make_identifier(s)


def evidence_for(qctx, ctx, evctx=None):
    """
    Returns an iterable of Evidence

    Parameters
    ----------
    qctx : object
        an object supported by evidence. If the object is a
        :class:`~owmeta.context.Context` with no identifier, then the query
        considers statements 'staged' (rather than stored) in the context
    ctx : Context
        Context that bounds where we look for statements about `qctx`. The
        contexts for statements found in this context are the actual targets of
        Evidence.supports statements.
    evctx : Context
        if the Evidence.supports statements should be looked for somewhere other
        than `ctx`, that can be specified in evctx. optional
    """
    if not evctx:
        evctx = ctx
    ctxs = query_context(ctx.rdf_graph(), qctx)
    ev_objs = []
    for c in ctxs:
        ev = evctx(Evidence)()
        ev.supports(Context(ident=c.identifier).rdf_object)
        for x in ev.load():
            ev_objs.append(x)
    return ev_objs


def query_context(graph, qctx):
    '''
    graph : rdflib.graph.Graph
        Graph where we can find the contexts for statements in `qctx`
    qctx : owmeta.context.Context
        Container for statements
    '''
    trips = qctx.contents_triples()
    lctx = None
    for t in trips:
        ctxs = graph.contexts(t)
        if lctx is None:
            lctx = frozenset(ctxs)
            continue
        if len(lctx) == 0:
            return frozenset()
        else:
            lctx = frozenset(ctxs) & lctx
            if len(lctx) == 0:
                return lctx
    return frozenset() if lctx is None else lctx



import six
from owmeta_core.contextualize import contextualize_metaclass, contextualized_new
from owmeta_core.context import Context, ContextMeta


class DocumentContextMeta(ContextMeta):

    def contextualize_class(self, context):
        if context is None:
            return self
        return contextualize_metaclass(context, self)('_H2', (self,), dict())


class DocumentContext(six.with_metaclass(DocumentContextMeta, Context)):
    """ A Context that corresponds to a document. """

    def __init__(self, document):
        super(DocumentContext, self).__init__()
        self._document = document

    @property
    def identifier(self):
        return self._document.make_identifier(self._document.identifier.n3())

    @identifier.setter
    def identifier(self, v):
        pass


DocumentContext.__new__ = contextualized_new(DocumentContext)



import six

from owmeta_core.command_util import GeneratorWithData
from ..cell import Cell


class CellCmd(object):
    '''
    Commands for dealing with biological cells
    '''

    def __init__(self, parent, *args, **kwargs):
        self._parent = parent

    def show(self, cell_name_or_id, context=None):
        '''
        Show information about the cell

        Parameters
        ----------
        cell_name_or_id : str
            Cell name or Cell URI
        context : str
            Context to search in. Optional, defaults to the default context.
        '''

        if not context:
            ctx = self._parent._default_ctx
        else:
            ctx = Context(context)

        def helper():
            for cell in ctx.stored(Cell).query(name=cell_name_or_id).load():
                yield cell
                break
            else: # no break
                uri = self._parent._den3(cell_name_or_id)
                for x in ctx.stored(Cell)(ident=uri).load():
                    yield cell

        def fmt_text(cell):
            try:
                sio = six.StringIO()

                nm = self._parent._conf('rdf.namespace_manager')
                print('%s(%s)' % (cell.__class__.__name__, nm.normalizeUri(cell.identifier)), file=sio)
                print('    Lineage name: %s' % cell.lineageName(), file=sio)
                print('    Name: %s' % cell.name(), file=sio)
                print('    WormBase ID: %s' % cell.wormbaseID(), end='', file=sio)
                return sio.getvalue()
            except AttributeError:
                res = str(cell)
                L.error('Failed while creating formatting string representation for %s', res)
                return res

        return GeneratorWithData(helper(), text_format=fmt_text)



import neuroml as N
from owmeta_core.data import DataUser
from owmeta_core.dataobject import DataObject, DatatypeProperty
import owmeta_core.dataobject_property as DP
from rdflib.term import URIRef
from rdflib.namespace import Namespace

from . import CONTEXT, BASE_BIO_SCHEMA_URL

NEUROML_NS = Namespace(f'{BASE_BIO_SCHEMA_URL}/NeuroML#')


class NeuroMLDocument(DataObject):
    '''
    Represents a NeuroML document

    The document may be represented literally in the RDF graph using `xml_content` or
    stored elsewhere and included by reference with `document_url`.

    Example::

        >>> embedded_nml = NeuroMLDocument(key='embedded_ex', content="""\\
        ... <?xml version="1.0" encoding="UTF-8"?>
        ... <neuroml xmlns="http://www.neuroml.org/schema/neuroml2"
        ...     xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
        ...     xsi:schemaLocation="http://www.neuroml.org/schema/neuroml2
        ...     https://raw.github.com/NeuroML/NeuroML2/master/Schemas/NeuroML2/NeuroML_v2beta.xsd"
        ...     id="k_slow">
        ...     <ionChannel id="k_slow" conductance="10pS" type="ionChannelHH" species="k">
        ...         <notes>K slow channel from Boyle and Cohen 2008</notes>
        ...         <gateHHtauInf id="n" instances="1">
        ...             <timeCourse type="fixedTimeCourse" tau="25.0007 ms"/>
        ...             <steadyState type="HHSigmoidVariable" rate="1" scale="15.8512 mV" midpoint="19.8741 mV"/>
        ...         </gateHHtauInf>
        ...     </ionChannel>
        ... </neuroml>""")

        >>> external_nml = NeuroMLDocument(ident='external_ex',
        ...     document_url='')

    '''

    class_context = CONTEXT

    content = DatatypeProperty()
    '''
    XML content for the document. Should be a complete NeuroML document rather than a
    fragment.
    '''

    document_url = DatatypeProperty(multiple=True)
    '''
    URL where the XML content of the document can be retrieved
    '''


class NeuroMLProperty(DP.ObjectProperty):
    '''
    Property for attaching NeuroML documents to resources
    '''
    class_context = CONTEXT
    link = NEUROML_NS['neuroML']
    value_type = NeuroMLDocument



import csv
import logging

from owmeta_core.datasource import GenericTranslation
from owmeta_core.dataobject import ObjectProperty
from owmeta_core.data_trans.csv_ds import CSVDataTranslator, CSVDataSource

from .. import CONTEXT
from ..utils import normalize_cell_name
from ..connection import Connection
from ..cell import Cell
from ..document import Document
from ..evidence import Evidence
from ..neuron import Neuron
from ..muscle import Muscle, BodyWallMuscle
from ..worm import Worm
from ..network import Network

from .common_data import DSMixin, DTMixin
from .data_with_evidence_ds import DataWithEvidenceDataSource

L = logging.getLogger(__name__)


class ConnectomeCSVDataSource(DSMixin, CSVDataSource):
    '''
    A CSV data source whose CSV file describes a neural connectome

    Basically, this is just a marker type to indicate what's described in the CSV --
    there's no consistent schema
    '''
    class_context = CONTEXT


class NeuronConnectomeCSVTranslation(GenericTranslation):
    class_context = CONTEXT

    neurons_source = ObjectProperty(value_type=DataWithEvidenceDataSource)
    muscles_source = ObjectProperty(value_type=DataWithEvidenceDataSource)

    key_properties = (GenericTranslation.source, muscles_source, neurons_source)


class NeuronConnectomeCSVTranslator(DTMixin, CSVDataTranslator):
    class_context = CONTEXT

    input_type = (ConnectomeCSVDataSource, DataWithEvidenceDataSource)
    output_type = DataWithEvidenceDataSource
    translation_type = NeuronConnectomeCSVTranslation

    def make_translation(self, sources):
        tr = super(NeuronConnectomeCSVTranslator, self).make_translation()
        tr.source(sources[0])
        tr.neurons_source(sources[1])
        tr.muscles_source(sources[2])
        return tr

    def translate(self, data_source, neurons_source, muscles_source):
        # muscle cells that are generically defined in source and need to be broken
        # into pair of L and R before being added to owmeta

        # muscle cells that have different names in connectome source and cell list.
        # Their wormbase cell list names will be used in owmeta
        changed_muscles = ['ANAL', 'INTR', 'INTL', 'SPH']

        # counters for terminal printing
        neuron_connections = 0
        muscle_connections = 0
        other_connections = 0

        res = self.make_new_output(sources=(data_source, neurons_source, muscles_source))

        # XXX: should there be a different src for muscles?
        w_q = muscles_source.data_context.stored(Worm)()
        n_q = neurons_source.data_context.stored(Network)()
        i_n = next(n_q.load(), n_q)
        i_w = next(w_q.load(), w_q)
        # XXX: If the context we query for is the same ID as the default
        # context, then we should query the merge of all graphs (using the
        # contract of the ConjunctiveGraph with 'default_is_union==True').
        # Otherwise, we should query only for what's in that graph.
        #
        neuron_objs = list(set(i_n.neurons()))
        muscle_objs = list(i_w.muscles())
        # get lists of neuron and muscles names
        neurons = set(neuron.name() for neuron in neuron_objs)
        muscles = set(muscle.name() for muscle in muscle_objs)
        o_w = res.data_context(Worm)()
        o_n = res.data_context(Network)(worm=o_w)
        # Evidence object to assert each connection
        ctxd_Document = res.evidence_context(Document)
        doc = ctxd_Document(key="emmons2015",
                            title='Whole-animal C. elegans connectomes',
                            year=2015,
                            uri='http://abstracts.genetics-gsa.org/cgi-bin/'
                            'celegans15s/wsrch15.pl?author=emmons&sort=ptimes&'
                            'sbutton=Detail&absno=155110844&sid=668862',
                            rdfs_comment="Data related by personal communication")
        doc.author('Emmons, S.')
        doc.author('Cook, S.')
        doc.author('Jarrell, T.')
        doc.author('Wang, Y.')
        doc.author('Yakolev, M.')
        doc.author('Nguyen, K.')
        doc.author('Hall, D.')
        docctx = res.data_context_for(document=doc)
        res.evidence_context(Evidence)(key="emmons2015",
                reference=doc,
                supports=docctx.rdf_object)
        with docctx(Neuron, Muscle, BodyWallMuscle, Cell, Connection) as ctx:
            res.data_context.add_import(ctx.context)
            with open(data_source.full_path()) as csvfile:
                edge_reader = csv.reader(csvfile)
                next(edge_reader)  # skip header row
                for row in edge_reader:
                    source, target, weight, syn_type = map(str.strip, row)

                    # set synapse type to something the Connection object
                    # expects, and normalize the source and target names
                    if syn_type == 'electrical':
                        syn_type = 'gapJunction'
                    elif syn_type == 'chemical':
                        syn_type = 'send'

                    source_is_bwm = 'BWM' in source.upper()
                    target_is_bwm = 'BWM' in target.upper()

                    source = normalize_cell_name(source).upper()
                    target = normalize_cell_name(target).upper()

                    weight = int(weight)

                    # change certain muscle names to names in wormbase
                    if source in changed_muscles:
                        source = changed_muscle(source)
                    if target in changed_muscles:
                        target = changed_muscle(target)

                    sources = convert_to_cell(ctx, source, muscles, neurons, source_is_bwm)
                    targets = convert_to_cell(ctx, target, muscles, neurons, target_is_bwm)

                    for s in sources:
                        for t in targets:
                            conn = add_synapse(ctx, s, t, weight, syn_type)
                            o_n.synapse(conn)
                            kind = conn.termination.onedef()
                            if kind == 'muscle':
                                muscle_connections += 1
                            elif kind == 'neuron':
                                neuron_connections += 1
                            else:
                                other_connections += 1

        print('Total neuron to neuron connections added = %i' % neuron_connections)
        print('Total neuron to muscle connections added = %i' % muscle_connections)
        print('Total other connections added = %i' % other_connections)
        print('uploaded connections')
        return res


class NeuronConnectomeSynapseClassTranslation(GenericTranslation):
    class_context = CONTEXT

    neurotransmitter_source = ObjectProperty()

    key_properties = (GenericTranslation.source, neurotransmitter_source)


class NeuronConnectomeSynapseClassTranslator(DTMixin, CSVDataTranslator):
    '''
    Adds synapse classes to existing connections
    '''
    class_context = CONTEXT

    translation_type = NeuronConnectomeSynapseClassTranslation

    input_type = (DataWithEvidenceDataSource, ConnectomeCSVDataSource)
    output_type = DataWithEvidenceDataSource

    def make_translation(self, sources):
        tr = super(NeuronConnectomeSynapseClassTranslator, self).make_translation()
        tr.source(sources[0])
        tr.neurotransmitter_source(sources[1])
        return tr

    def translate(self, data_source, neurotransmitter_source):
        res = self.make_new_output(sources=(data_source, neurotransmitter_source))
        doc = res.evidence_context(Document)(author=['Dimitar Sht. Shterionov', 'Gerda Janssens'],
                       title='Data acquisition and modeling for learning and'
                             ' reasoning in probabilistic logic environment',
                       year=2011,
                       uri='https://groups.google.com/forum/m/#!topic/openworm-discuss/G9wKoR8N-l0/discussion')
        e = res.evidence_context(Evidence)(key="shterionov2011", reference=doc)

        # We don't use doc.as_context for the statements' context because the document
        # itself doesn't actually make the statements below, although it does *support*
        # them by describing the process by which they are derived
        docctx = res.data_context_for(document=doc)

        # There are many entries in the Shterionov data where the number of synapses
        # doesn't match that in, say, the Emmons (2015) data. We distinguish these so it's
        # easy to look back at the differences later, but the different number of synapses
        # isn't recorded here.
        docctx_anynum = res.data_context_for(document=doc, number_matches=False)

        e.supports(docctx.rdf_object)
        res.data_context.add_import(docctx)
        res.data_context.add_import(docctx_anynum)
        with data_source.data_context.stored(Connection, Neuron) as srcctx:
            with self.make_reader(neurotransmitter_source,
                                  skipheader=False,
                                  delimiter=';') as reader:
                for row in reader:
                    pre, post, typ, number, nt = row
                    conn = srcctx.Connection.query(pre_cell=srcctx.Neuron.query(pre),
                                                   post_cell=srcctx.Neuron.query(post),
                                                   number=int(number),
                                                   syntype=typ)
                    hit = False
                    for c in conn.load():
                        docctx(Connection)(ident=c.identifier).synclass(nt)
                        hit = True

                    if not hit:
                        conn = srcctx.Connection.query(pre_cell=srcctx.Neuron(pre),
                                                       post_cell=srcctx.Neuron(post),
                                                       syntype=typ)
                        hit = False
                        for c in conn.load():
                            docctx_anynum(Connection)(ident=c.identifier).synclass(nt)
                            hit = True

                        if not hit:
                            L.warning("Didn't find any connections matching: {}".format(conn))
        return res


def convert_to_cell(ctx, name, muscles, neurons, is_bwm):
    ret = []
    res = None
    res2 = None
    if name in neurons:
        res = ctx.Neuron(name)
    elif name in TO_EXPAND_MUSCLES:
        # All of the muscle cell types we need to expand are pharyngeal, so no need to
        # pass through is_bwm
        res, res2 = expand_muscle(ctx, name)
    elif name in muscles:
        if is_bwm:
            res = ctx.BodyWallMuscle(name)
        else:
            res = ctx.Muscle(name)
    elif name in OTHER_CELLS:
        res = ctx.Cell(name)

    if res is not None:
        ret.append(res)
    if res2 is not None:
        ret.append(res2)

    return ret


def add_synapse(ctx, source, target, weight, syn_type):
    c = ctx.Connection(pre_cell=source, post_cell=target,
                       number=weight, syntype=syn_type)

    if isinstance(source, Neuron) and isinstance(target, Neuron):
        c.termination('neuron')
    elif isinstance(source, Neuron) and isinstance(target, Muscle) or \
            isinstance(source, Muscle) and isinstance(target, Neuron):
        c.termination('muscle')

    return c


MUSCLES = {
    'ANAL': 'MU_ANAL',
    'INTR': 'MU_INT_R',
    'INTL': 'MU_INT_L',
    'SPH': 'MU_SPH'
}

TO_EXPAND_MUSCLES = ['PM1D', 'PM2D', 'PM3D', 'PM4D', 'PM5D']

#
# cells that are neither neurons or muscles. These are marked as
# 'Other Cells' in the wormbase cell list but are still part of the new
# connectome.
#
# TODO: In future work these should be uploaded seperately to
# owmeta in a new upload function and should be referred from there
# instead of this list.
OTHER_CELLS = ['MC1DL', 'MC1DR', 'MC1V', 'MC2DL', 'MC2DR', 'MC2V', 'MC3DL',
               'MC3DR', 'MC3V']


def changed_muscle(x):
    return MUSCLES[x]


def expand_muscle(ctx, name):
    return ctx(Muscle)(name + 'L'), ctx(Muscle)(name + 'R')



from rdflib.namespace import Namespace
import csv
from owmeta_core.datasource import Informational
from owmeta_core.data_trans.csv_ds import CSVDataSource, CSVDataTranslator
import re
import traceback


from .. import CONTEXT
from ..utils import normalize_cell_name
from ..channel import Channel, ExpressionPattern
from ..evidence import Evidence
from ..muscle import Muscle, BodyWallMuscle
from ..network import Network
from ..neuron import Neuron
from ..cell import Cell
from ..website import Website
from ..worm import Worm

from .common_data import DSMixin, DTMixin
from .data_with_evidence_ds import DataWithEvidenceDataSource


class WormbaseTextMatchCSVDataSource(DSMixin, CSVDataSource):
    class_context = CONTEXT

    initial_cell_column = Informational('Initial Cell Column',
                                        description='The index of the first column with a cell name',
                                        multiple=False)

    cell_type = Informational('Cell Type',
                              description='The type of cell to be produced',
                              multiple=False)


class WormbaseIonChannelCSVDataSource(DSMixin, CSVDataSource):
    class_context = CONTEXT

    csv_header = ['channel_name',
                  'gene_name',
                  'gene_WB_ID',
                  'expression_pattern',
                  'description']


class WormbaseIonChannelCSVTranslator(DTMixin, CSVDataTranslator):
    class_context = CONTEXT

    input_type = WormbaseIonChannelCSVDataSource
    output_type = DataWithEvidenceDataSource

    def translate(self, data_source):
        res = self.make_new_output((data_source,))
        with res.evidence_context(Evidence=Evidence, Website=Website) as ctx:
            doc = ctx.Website(key="wormbase", url="http://Wormbase.org", title="WormBase")
            doc_ctx = res.data_context_for(document=doc)
            ctx.Evidence(reference=doc, supports=doc_ctx.rdf_object)

        with self.make_reader(data_source) as csvreader:
            with doc_ctx(Channel=Channel,
                         ExpressionPattern=ExpressionPattern) as ctx:
                for line in csvreader:
                    channel_name = normalize_cell_name(line[0]).upper()
                    gene_name = line[1].upper()
                    gene_WB_ID = line[2].upper()
                    expression_pattern = line[3]
                    description = line[4]
                    c = ctx.Channel(name=str(channel_name))
                    c.gene_name(gene_name)
                    c.gene_WB_ID(gene_WB_ID)
                    c.description(description)
                    patterns = expression_pattern.split(r' | ')
                    regex = re.compile(r' *\[([^\]]+)\] *(.*) *')

                    matches = [regex.match(pat) for pat in patterns]
                    patterns = [ctx.ExpressionPattern(wormbaseid=m.group(1),
                                                      description=m.group(2))
                                for m in matches if m is not None]
                    for pat in patterns:
                        c.expression_pattern(pat)
        return res


class WormbaseTextMatchCSVTranslator(DTMixin, CSVDataTranslator):
    class_context = CONTEXT

    input_type = WormbaseTextMatchCSVDataSource
    output_type = DataWithEvidenceDataSource

    def translate(self, data_source):
        initcol = data_source.initial_cell_column()
        ctype = data_source.cell_type()

        ctype = self.context.resolve_class(ctype)

        res = self.make_new_output((data_source,))
        with res.evidence_context(Evidence=Evidence, Website=Website) as ctx:
            doc = ctx.Website(key="wormbase", url="http://Wormbase.org", title="WormBase")
            doc_ctx = res.data_context_for(document=doc)
            ctx.Evidence(reference=doc, supports=doc_ctx.rdf_object)

        with open(data_source.full_path(), 'r') as f:
            reader = csv.reader(f, delimiter='\t')
            header = self.skip_to_header(reader)
            with doc_ctx(Channel=Channel, CType=ctype) as ctx:
                for row in reader:
                    cells = self.extract_cell_names(header,
                                                    initcol,
                                                    row)
                    ch = ctx.Channel(name=str(row[0]))
                    for cell in cells:
                        m = ctx.CType(name=str(cell))
                        ch.appearsIn(m)
        return res

    def skip_to_header(self, reader):
        rows = 0
        for row in reader:
            if rows == 3:
                return row
            rows += 1
        return None

    def extract_cell_names(self, header, initial_cell_column, row):
        res = []
        cols = 0
        for col in row:
            if cols > initial_cell_column:
                if col == '1' or col == '2':
                    res.append(header[cols])
            cols += 1
        return res


class WormBaseCSVDataSource(DSMixin, CSVDataSource):
    class_context = CONTEXT

    csv_header = ["Cell",
                  "Lineage Name",
                  "Description",
                  "Total count of identified adult-only hermaphrodite cells",
                  "Total count of adult-only male cells",
                  "Neurons (no male-specific cells)",
                  "Neurons (male-specific)",
                  "Body wall muscles",
                  "Pharynx muscles",
                  "Other muscles",
                  "Other adult-only cells in the hermaphrodite",
                  "Other adult-only hermaphrodite-specific cells (not present in males)",
                  "Motor neurons related to body wall muscles",
                  "Embryonic cells not present in adult",
                  "Male-specific cells",
                  "Male-specific adult-only cells",
                  "Cells with non-unique name",
                  "",
                  "VirtualWorm blender model names",
                  "WormBase ID",
                  "Synonyms"]


class CellWormBaseCSVTranslator(DTMixin, CSVDataTranslator):
    class_context = CONTEXT

    input_type = WormBaseCSVDataSource
    output_type = DataWithEvidenceDataSource

    def translate(self, data_source):
        """ Translate wormbase CSV dump into Cells, Neurons, and Muscles """
        res = self.make_new_output((data_source,))
        with self.make_reader(data_source, skipheader=False, skiplines=3,
                dict_reader=True,
                fieldnames=data_source.csv_header.one()) as csvreader:
            # TODO: Improve this evidence by going back to the actual research
            #       by using the wormbase REST API in addition to or instead of the CSV file
            with res.evidence_context(Evidence=Evidence, Website=Website) as ctx:
                doc = ctx.Website(key="wormbase", url="http://Wormbase.org", title="WormBase")
                doc_ctx = res.data_context_for(document=doc)
                ctx.Evidence(reference=doc, supports=doc_ctx.rdf_object)

            with doc_ctx(Worm=Worm,
                         BodyWallMuscle=BodyWallMuscle,
                         Muscle=Muscle,
                         Network=Network,
                         Neuron=Neuron,
                         Cell=Cell) as ctx:
                w = ctx.Worm()
                n = ctx.Network()
                n.worm(w)

                for line in csvreader:
                    cell = None
                    if line['Body wall muscles']:
                        cell = ctx.BodyWallMuscle()
                        w.muscle(cell)
                    elif line['Pharynx muscles'] or line['Other muscles']:
                        cell = ctx.Muscle()
                        w.muscle(cell)
                    elif line['Neurons (no male-specific cells)']:
                        cell = ctx.Neuron()
                        cell.wormbaseID(line['WormBase ID'])
                        n.neuron(cell)
                    elif (line['Other adult-only cells in the hermaphrodite'] or
                            line['Other adult-only hermaphrodite-specific cells (not present in males)']):
                        cell = ctx.Cell()

                    if cell:
                        cell.wormbaseID(line['WormBase ID'])
                        cell.name(normalize_cell_name(line['Cell']).upper())
                        cell.description(line['Description'])
                        w.cell(cell)
        return res



from __future__ import print_function

from wrapt import ObjectProxy
from owmeta_core.custom_dataobject_property import CustomProperty
from owmeta_core.dataobject import DatatypeProperty, Alias

from .cell import Cell
from .connection import Connection


class NeuronProxy(ObjectProxy):

    def __init__(self, neighbor, connection, *args):
        super(NeuronProxy, self).__init__(*args)
        self._self_neighbor = neighbor
        self._self_connection = connection

    @property
    def neighbor(self):
        return self._self_neighbor

    @property
    def connection(self):
        return self._self_connection


class Neuron(Cell):
    """
    A neuron.

    See what neurons express some neuropeptide

    Example::

        # Grabs the representation of the neuronal network
        >>> net = P.Worm().get_neuron_network()

        # Grab a specific neuron
        >>> aval = net.aneuron('AVAL')

        >>> aval.type()
        set([u'interneuron'])

        #show how many connections go out of AVAL
        >>> aval.connection.count('pre')
        77

        >>> aval.name()
        u'AVAL'

        #list all known receptors
        >>> sorted(aval.receptors())
        [u'GGR-3', u'GLR-1', u'GLR-2', u'GLR-4', u'GLR-5', u'NMR-1', u'NMR-2', u'UNC-8']

        #show how many chemical synapses go in and out of AVAL
        >>> aval.Syn_degree()
        90

    Parameters
    ----------
    name : string
        The name of the neuron.

    Attributes
    ----------
    neighbor : CustomProperty
        Get neurons connected to this neuron if called with no arguments, or
        with arguments, state that neuronName is a neighbor of this Neuron
    connection : CustomProperty
        Get a set of Connection objects describing chemical synapses or gap
        junctions between this neuron and others

    """

    class_context = Cell.class_context

    type = DatatypeProperty(multiple=True)
    ''' The neuron type (i.e., sensory, interneuron, motor) '''

    receptor = DatatypeProperty(multiple=True)
    ''' The receptor types associated with this neuron '''

    innexin = DatatypeProperty(multiple=True)
    ''' Innexin types associated with this neuron '''

    neurotransmitter = DatatypeProperty(multiple=True)
    ''' Neurotransmitters associated with this neuron '''

    neuropeptide = DatatypeProperty(multiple=True)
    ''' Name of the gene corresponding to the neuropeptide produced by this neuron '''

    receptors = Alias(receptor)
    ''' Alias to :py:attr:`receptor` '''

    def __init__(self, name=None, **kwargs):
        super(Neuron, self).__init__(name=name, **kwargs)

        # Get neurons connected to this neuron
        Neighbor(owner=self)

        # Get connections from this neuron
        ConnectionProperty(owner=self)

        self.get_neighbors = self.neighbor

    def contextualize(self, context):
        res = super(Neuron, self).contextualize(context)
        if hasattr(self, 'neighbor'):
            res = NeuronProxy(self.neighbor.contextualize(context),
                              self.connection.contextualize(context),
                              res)
        return res

    def GJ_degree(self):
        """Get the degree of this neuron for gap junction edges only

        :returns: total number of incoming and outgoing gap junctions
        :rtype: int
        """
        count = 0
        for c in self.connection():
            if c.syntype.one() == 'gapJunction':
                count += 1
        return count

    def Syn_degree(self):
        """Get the degree of this neuron for chemical synapse edges only

        :returns: total number of incoming and outgoing chemical synapses
        :rtype: int
        """
        count = 0
        for c in self.connection.get('either'):
            if c.syntype.one() == 'send':
                count += 1
        return count

    def get_incidents(self, type=0):
        """ Get neurons which synapse at this neuron """
        # Directed graph. Getting accessible _from_ this node
        for item in self['nx'].in_edges_iter(self.name(), data=True):
            if 'GapJunction' in item[2]['synapse']:
                yield item[0]


class Neighbor(CustomProperty):
    multiple = True

    def __init__(self, **kwargs):
        super(Neighbor, self).__init__('neighbor', **kwargs)
        self._conns = []
        self._conntype = Connection.contextualize(self.owner.context)
        self.context = self.owner.context

    def contextualize(self, context):
        res = type(self)(owner=self.owner)
        res._conns = self._conns
        res._conntype = Connection.contextualize(context)
        res.context = context
        return res

    def get(self, **kwargs):
        """Get a list of neighboring neurons.

           Parameters
           ----------
           See parameters for owmeta.connection.Connection

           Returns
           -------
           list of Neuron
        """
        if len(self._conns) > 0:
            for c in self._conns:
                if c.context == self.context:
                    for post in c.post_cell.get():
                        yield post
        else:
            conn = self._conntype.contextualize(self.context)(pre_cell=self.owner, **kwargs)
            for r in conn.post_cell.get():
                yield r

    def get_terms(self, **kwargs):
        """Get a list of neighboring neurons.

           Parameters
           ----------
           See parameters for owmeta.connection.Connection

           Returns
           -------
           list of Neuron
        """
        if len(self._conns) > 0:
            for c in self._conns:
                if c.context == self.context:
                    for post in c.post_cell.get_terms():
                        yield post
        else:
            conn = self._conntype.contextualize(self.context)(pre_cell=self.owner, **kwargs)
            for r in conn.post_cell.get_terms():
                yield r

    def count(self, **kwargs):
        conntype = self._conntype.contextualize(self.context)
        return conntype(pre_cell=self.owner, **kwargs).count()

    @property
    def defined_values(self):
        return []

    @property
    def values(self):
        return []

    def set(self, other, **kwargs):
        c = self._conntype(pre_cell=self.owner, post_cell=other, **kwargs)
        self._conns.append(c)
        return c

    def triples(self, **kwargs):
        for c in self._conns:
            for x in c.triples(**kwargs):
                yield x


class ConnectionProperty(CustomProperty):

    """A representation of the connection between neurons. Either a gap junction
    or a chemical synapse

    TODO: Add neurotransmitter type.
    TODO: Add connection strength
    """

    multiple = True

    def __init__(self, **kwargs):
        super(ConnectionProperty, self).__init__('connection', **kwargs)
        self._conns = []
        self._conntype = Connection.contextualize(self.owner.context)

    def get(self, pre_post_or_either='pre', **kwargs):
        """Get a list of connections associated with the owning neuron.

           Parameters
           ----------
           pre_post_or_either: str
               What kind of connection to look for.
               'pre': Owner is the source of the connection
               'post': Owner is the destination of the connection
               'either': Owner is either the source or destination of the connection

           Returns
           -------
           list of Connection
        """
        c = self._gather_query_conns(pre_post_or_either, **kwargs)

        for x in c:
            for r in x.load():
                yield r

        for x in self._conns:
            if x.defined and x.context == self.context:
                yield x

    def get_terms(self, pre_post_or_either='pre', **kwargs):
        """
        Get a list of connection identifiers associated with the owning neuron.

        Parameters
        ----------
        pre_post_or_either: str
            What kind of connection to look for.
            'pre': Owner is the source of the connection
            'post': Owner is the destination of the connection
            'either': Owner is either the source or destination of the connection

        Returns
        -------
        list of Connection
        """
        c = self._gather_query_conns(pre_post_or_either, **kwargs)

        for x in c:
            for r in x.load_terms():
                yield r

        for x in self._conns:
            if x.defined and x.context == self.context:
                yield x.identifier

    def _gather_query_conns(self, pre_post_or_either, **kwargs):
        c = []
        ct = self._conntype.contextualize(self.context)
        if pre_post_or_either == 'pre':
            c.append(ct(pre_cell=self.owner, **kwargs))
        elif pre_post_or_either == 'post':
            c.append(ct(post_cell=self.owner, **kwargs))
        elif pre_post_or_either == 'either':
            c.append(ct(pre_cell=self.owner, **kwargs))
            c.append(ct(post_cell=self.owner, **kwargs))
        return c

    def contextualize(self, context):
        res = type(self)(owner=self.owner)
        res._conns = self._conns
        res.context = context
        return res

    @property
    def values(self):
        return []

    def count(self, pre_post_or_either='pre', *args, **kwargs):
        res = 0
        conntype = self._conntype.contextualize(self.context)
        if pre_post_or_either == 'pre':
            res += conntype(pre_cell=self.owner, **kwargs).count()
        elif pre_post_or_either == 'post':
            res += conntype(post_cell=self.owner, **kwargs).count()
        elif pre_post_or_either == 'either':
            res += conntype(pre_cell=self.owner, **kwargs).count() + \
                    conntype(post_cell=self.owner, **kwargs).count()

        return res

    def set(self, conn, **kwargs):
        """Add a connection associated with the owner Neuron

           Parameters
           ----------
           conn : owmeta.connection.Connection
               connection associated with the owner neuron

           Returns
           -------
           A owmeta.neuron.Connection
        """
        assert(isinstance(conn, self._conntype))
        self._conns.append(conn)

    def triples(self, **kwargs):
        for c in self._conns:
            for x in c.triples(**kwargs):
                yield x



from .. import BASE_SCHEMA_URL, BASE_DATA_URL
from rdflib.namespace import Namespace

TRANS_NS = Namespace(BASE_SCHEMA_URL + '/translators/')
TRANS_DATA_NS = Namespace(BASE_DATA_URL + '/translators/')
DS_NS = Namespace(BASE_SCHEMA_URL + '/data_sources/')
DS_DATA_NS = Namespace(BASE_DATA_URL + '/data_sources/')


class DSMixin(object):
    base_namespace = DS_NS
    base_data_namespace = DS_DATA_NS


class DTMixin(object):
    base_namespace = TRANS_NS
    base_data_namespace = TRANS_DATA_NS



from owmeta_core.datasource import DataTranslator, OneOrMore

from .. import SCI_CTX

from .common_data import DTMixin
from .data_with_evidence_ds import DataWithEvidenceDataSource


class ContextMergeDataTranslator(DTMixin, DataTranslator):
    class_context = SCI_CTX

    input_type = OneOrMore(DataWithEvidenceDataSource)
    output_type = DataWithEvidenceDataSource

    def translate(self, *sources):
        if not sources:
            raise Exception("No sources were provided")

        sources = sorted(sources, key=lambda s: s.identifier)
        res = self.make_new_output(sources=sources)

        for src in sources:
            res.data_context.add_import(src.data_context)
            res.evidence_context.add_import(src.evidence_context)
        return res



import rdflib as R

from .biology import BiologyType
from .cell_common import CELL_RDF_TYPE
from .channel_common import CHANNEL_RDF_TYPE
from .channelworm import ChannelModel

from owmeta_core.dataobject import DatatypeProperty, ObjectProperty, Alias


class ExpressionPattern(BiologyType):

    class_context = BiologyType.class_context

    wormbaseid = DatatypeProperty()
    ''' The ID for the expression pattern in Wormbase '''

    wormbaseID = Alias(wormbaseid)
    ''' Alias to :py:attr:`wormbaseid` '''

    wormbaseURL = DatatypeProperty()
    ''' The URL for the expression pattern in Wormbase '''

    description = DatatypeProperty()
    ''' Natural language description of the expression pattern '''

    def __init__(self, wormbaseid=None, **kwargs):
        super(ExpressionPattern, self).__init__(**kwargs)

        if wormbaseid:
            self.wormbaseid(wormbaseid)
            self.wormbaseURL(R.URIRef("http://www.wormbase.org/species/all/expr_pattern/" + wormbaseid))

    def defined_augment(self):
        return self.wormbaseid.has_defined_value()

    def identifier_augment(self):
        return self.make_identifier(self.wormbaseid.defined_values[0])


class Channel(BiologyType):
    """
    A biological ion channel.
    """

    class_context = BiologyType.class_context

    rdf_type = CHANNEL_RDF_TYPE

    subfamily = DatatypeProperty()
    ''' Ion channel's subfamily '''

    name = DatatypeProperty()
    ''' Ion channel's name '''

    description = DatatypeProperty()
    ''' A description of the ion channel '''

    gene_name = DatatypeProperty()
    ''' Name of the gene that codes for this ion channel '''

    gene_class = DatatypeProperty()
    ''' Classification of the encoding gene '''

    gene_WB_ID = DatatypeProperty()
    ''' Wormbase ID of the encoding gene '''

    expression_pattern = ObjectProperty(multiple=True, value_type=ExpressionPattern)
    ''' A pattern of expression of this cell within an organism '''

    proteins = DatatypeProperty(multiple=True)
    ''' Proteins associated with this channel '''

    appearsIn = ObjectProperty(multiple=True, value_rdf_type=CELL_RDF_TYPE)
    ''' Cell types in which the ion channel has been expressed '''

    model = ObjectProperty(value_type=ChannelModel)
    ''' Get experimental models of this ion channel '''

    models = Alias(model)
    ''' Alias to :py:attr:`model` '''

    def __init__(self, name=None, **kwargs):
        super(Channel, self).__init__(name=name, **kwargs)

    def defined_augment(self):
        return self.name.has_defined_value()

    def identifier_augment(self):
        return self.make_identifier(self.name.defined_values[0])



"""
Common utilities for translation, massaging data, etc., that don't fit
elsewhere in owmeta
"""
import re

__all__ = ['normalize_cell_name']
# to normalize certain neuron and muscle names
SEARCH_STRING = re.compile(r'\w+0+[1-9]+')
REPLACE_STRING = re.compile(r'0+')
SEARCH_STRING_MUSCLE = re.compile(r'\w+BWM\w+')
REPLACE_STRING_MUSCLE = re.compile(r'BWM')


def normalize_cell_name(name):
    # normalize neuron and muscle names to match those used at other points
    # see #137 for elaboration
    # if there are zeroes in the middle of a name, remove them
    if re.match(SEARCH_STRING, name):
        name = REPLACE_STRING.sub('', name)
    name = normalize_muscle(name)
    name = name.upper()
    return name


def normalize_muscle(name):
    # normalize names of Body Wall Muscles
    # if there is 'BWM' in the name, remove it
    if re.match(SEARCH_STRING_MUSCLE, name):
        name = REPLACE_STRING_MUSCLE.sub('', name)
    return name



from owmeta_core.dataobject import DataObject, ObjectProperty

from . import SCI_CTX
from .document import Document


class Experiment(DataObject):
    """
    Generic class for storing information about experiments

    Should be overridden by specific types of experiments
    (example: see PatchClampExperiment in channelworm.py).

    Overriding classes should have a list called "conditions" that
    contains the names of experimental conditions for that particular
    type of experiment.
    Each of the items in "conditions" should also be either a
    DatatypeProperty or ObjectProperty for the experiment as well.
    """

    class_context = SCI_CTX

    reference = ObjectProperty(value_type=Document, multiple=True)
    ''' Supporting article for this experiment. '''

    def __init__(self, **kwargs):
        super(Experiment, self).__init__(**kwargs)
        self._condits = {}

    def get_conditions(self):
        """Return conditions and their associated values in a dict."""
        if not hasattr(self, 'conditions'):
            raise NotImplementedError('"conditions" attribute must be overridden')

        for c in self.conditions:
            value = getattr(self, c)
            if callable(value):
                self._condits[c] = value()
            else:
                if value:
                    #if property is not empty
                    self._condits[c] = value

        return self._condits



from __future__ import print_function, absolute_import

import logging

from owmeta_core.command_util import GenericUserError
from owmeta_core.dataobject import DataObject

from .commands.biology import CellCmd


L = logging.getLogger(__name__)


class OWMEvidence(object):
    '''
    Commands for evidence
    '''
    def __init__(self, parent):
        self._parent = parent

    def get(self, identifier, rdf_type=None):
        '''
        Retrieves evidence for the given object. If there are multiple types for the object, the evidence for only one
        type will be shown, but you can specify which type should be used.

        Parameters
        ----------
        identifier : str
            The object to show evidence for
        rdf_type : str
            Type of the object to show evidence
        '''
        from owmeta.evidence import Evidence
        from owmeta.data_trans.data_with_evidence_ds import DataWithEvidenceDataSource
        from owmeta_core.context_dataobject import ContextDataObject
        ctx = self._parent._default_ctx.stored
        identifier = self._parent._den3(identifier)
        rdf_type = self._parent._den3(rdf_type)
        if rdf_type:
            base_type = ctx(ctx.resolve_class(rdf_type))
            if base_type is None:
                raise GenericUserError("Could not find Python class corresponding to " +
                        str(rdf_type))
        else:
            base_type = ctx(DataObject)

        q = base_type.query(ident=identifier)
        for l in q.load():
            if isinstance(l, ContextDataObject):
                evq = ctx(Evidence).query()
                evq.supports(l)
                self._message_evidence(evq)
            if isinstance(l, DataWithEvidenceDataSource):
                evq = l.evidence_context.stored(Evidence).query()
                self._message_evidence(evq)

    def _message_evidence(self, evq):
        from owmeta.website import Website
        from owmeta.document import Document
        msg = self._parent.message
        for m in evq.load():
            ref = m.reference()
            if isinstance(ref, Document):
                msg(ref)
                titles = ['Author:',
                          'Title: ',
                          'URI:   ',
                          'DOI:   ',
                          'PMID:  ',
                          'WBID:  ']
                vals = [ref.author(),
                        ref.title(),
                        ref.uri(),
                        ref.doi(),
                        ref.pmid(),
                        ref.wbid()]
                for title, v in zip(titles, vals):
                    if v:
                        msg(title, v)
                msg()
            elif isinstance(ref, Website):
                msg(ref)
                titles = ['Title: ',
                          'URL:   ']
                vals = [ref.title(),
                        ref.url()]
                for title, v in zip(titles, vals):
                    if v:
                        msg(title, v)
                msg()



# -*- coding: utf-8 -*-
from __future__ import print_function

from owmeta_core.dataobject import ObjectProperty, Alias

from .connection import Connection
from .neuron import Neuron
from .biology import BiologyType
from .worm_common import WORM_RDF_TYPE


class Network(BiologyType):

    """ A network of neurons """

    class_context = BiologyType.class_context

    synapse = ObjectProperty(value_type=Connection, multiple=True)
    ''' Returns a set of all synapses in the network '''

    neuron = ObjectProperty(value_type=Neuron, multiple=True)
    ''' Returns a set of all Neuron objects in the network '''

    worm = ObjectProperty(value_rdf_type=WORM_RDF_TYPE)
    ''' The worm connected to the network '''

    synapses = Alias(synapse)
    ''' Alias to `synapse` '''

    neurons = Alias(neuron)
    ''' Alias to `neuron` '''

    def __init__(self, worm=None, **kwargs):
        super(Network, self).__init__(**kwargs)

        if worm is not None:
            self.worm(worm)

    def neuron_names(self):
        """
        Gets the complete set of neurons' names in this network.

        Example::

            # Grabs the representation of the neuronal network
            >>> net = Worm().get_neuron_network()

            >>> len(set(net.neuron_names()))
            302
            >>> set(net.neuron_names())
            set(['VB4', 'PDEL', 'HSNL', 'SIBDR', ... 'RIAL', 'MCR', 'LUAL'])

        """
        neuron_expr = self.neuron.expr
        neuron_expr.name()
        return set(x.toPython() for x in neuron_expr.name)

    def aneuron(self, name):
        """
        Get a neuron by name.

        Example::

            # Grabs the representation of the neuronal network
            >>> net = Worm().get_neuron_network()

            # Grab a specific neuron
            >>> aval = net.aneuron('AVAL')

            >>> aval.type()
            set([u'interneuron'])


        :param name: Name of a c. elegans neuron
        :returns: Neuron corresponding to the name given
        :rtype: owmeta.neuron.Neuron
        """
        return Neuron.contextualize(self.context)(name=name, conf=self.conf)

    def sensory(self):
        """
        Get all sensory neurons

        :returns: A iterable of all sensory neurons
        :rtype: iter(Neuron)
        """

        n = Neuron.contextualize(self.context)()
        n.type('sensory')

        self.neuron.set(n)
        res = list(n.load())
        self.neuron.unset(n)
        return res

    def interneurons(self):
        """
        Get all interneurons

        :returns: A iterable of all interneurons
        :rtype: iter(Neuron)
        """

        n = Neuron.contextualize(self.context)()
        n.type('interneuron')

        self.neuron.set(n)
        res = list(n.load())
        self.neuron.unset(n)
        return res

    def motor(self):
        """
        Get all motor

        :returns: A iterable of all motor neurons
        :rtype: iter(Neuron)
        """

        n = Neuron.contextualize(self.context)()
        n.type('motor')

        self.neuron.set(n)
        res = list(n.load())
        self.neuron.unset(n)
        return res

    def identifier_augment(self):
        return self.make_identifier(self.worm.defined_values[0].identifier.n3())

    def defined_augment(self):
        return self.worm.has_defined_value()
